"\n"
"General configuration for OpenCV 4.6.0 =====================================\n"
"  Version control:               4.6.0-dirty\n"
"\n"
"  Extra modules:\n"
"    Location (extra):            /workdir/git/contrib/modules\n"
"    Version control (extra):     4.6.0-11-g7b77c355a\n"
"\n"
"  Platform:\n"
"    Timestamp:                   2022-06-05T15:32:44Z\n"
"    Host:                        Linux 4.15.0-166-generic x86_64\n"
"    Target:                      Linux arm\n"
"    CMake:                       3.24.2\n"
"    CMake generator:             Ninja\n"
"    CMake build tool:            ninja\n"
"    Configuration:               Release\n"
"\n"
"  CPU/HW features:\n"
"    Baseline:                    NEON\n"
"      requested:                 DETECT\n"
"      disabled:                  VFPV3 NEON\n"
"\n"
"  C/C++:\n"
"    Built as dynamic libs?:      YES\n"
"    C++ standard:                11\n"
"    C++ Compiler:                /workdir/recipe-sysroot-native/usr/bin/arm-xilinx-linux-gnueabi/arm-xilinx-linux-gnueabi-g++  (ver 12.2.0)\n"
"    C++ flags (Release):         -mthumb -mfpu=neon -mfloat-abi=hard -mcpu=cortex-a9 -I/workdir/git/include  -fstack-protector-strong  -O2 -D_FORTIFY_SOURCE=2 -Wformat -Wformat-security -Werror=format-security  --sysroot=/workdir/recipe-sysroot  -O2 -pipe -g -feliminate-unused-debug-types -fmacro-prefix-map=/workdir/git=/usr/src/debug/opencv/4.6.0-r0  -fdebug-prefix-map=/workdir/git=/usr/src/debug/opencv/4.6.0-r0  -fmacro-prefix-map=/workdir/build=/usr/src/debug/opencv/4.6.0-r0  -fdebug-prefix-map=/workdir/build=/usr/src/debug/opencv/4.6.0-r0  -fdebug-prefix-map=/workdir/recipe-sysroot=  -fmacro-prefix-map=/workdir/recipe-sysroot=  -fdebug-prefix-map=/workdir/recipe-sysroot-native=  -fvisibility-inlines-hidden   -fsigned-char -W -Wall -Wreturn-type -Wnon-virtual-dtor -Waddress -Wsequence-point -Wformat -Wformat-security -Wmissing-declarations -Wundef -Winit-self -Wpointer-arith -Wshadow -Wsign-promo -Wuninitialized -Wsuggest-override -Wno-delete-non-virtual-dtor -Wno-comment -Wimplicit-fallthrough=3 -Wno-strict-overflow -fdiagnostics-show-option -pthread -fomit-frame-pointer -ffunction-sections -fdata-sections  -DNDEBUG  -DNDEBUG\n"
"    C++ flags (Debug):           -mthumb -mfpu=neon -mfloat-abi=hard -mcpu=cortex-a9 -I/workdir/git/include  -fstack-protector-strong  -O2 -D_FORTIFY_SOURCE=2 -Wformat -Wformat-security -Werror=format-security  --sysroot=/workdir/recipe-sysroot  -O2 -pipe -g -feliminate-unused-debug-types -fmacro-prefix-map=/workdir/git=/usr/src/debug/opencv/4.6.0-r0  -fdebug-prefix-map=/workdir/git=/usr/src/debug/opencv/4.6.0-r0  -fmacro-prefix-map=/workdir/build=/usr/src/debug/opencv/4.6.0-r0  -fdebug-prefix-map=/workdir/build=/usr/src/debug/opencv/4.6.0-r0  -fdebug-prefix-map=/workdir/recipe-sysroot=  -fmacro-prefix-map=/workdir/recipe-sysroot=  -fdebug-prefix-map=/workdir/recipe-sysroot-native=  -fvisibility-inlines-hidden   -fsigned-char -W -Wall -Wreturn-type -Wnon-virtual-dtor -Waddress -Wsequence-point -Wformat -Wformat-security -Wmissing-declarations -Wundef -Winit-self -Wpointer-arith -Wshadow -Wsign-promo -Wuninitialized -Wsuggest-override -Wno-delete-non-virtual-dtor -Wno-comment -Wimplicit-fallthrough=3 -Wno-strict-overflow -fdiagnostics-show-option -pthread -fomit-frame-pointer -ffunction-sections -fdata-sections  -g  -DDEBUG -D_DEBUG\n"
"    C Compiler:                  /workdir/recipe-sysroot-native/usr/bin/arm-xilinx-linux-gnueabi/arm-xilinx-linux-gnueabi-gcc\n"
"    C flags (Release):           -mthumb -mfpu=neon -mfloat-abi=hard -mcpu=cortex-a9 -I/workdir/git/include  -fstack-protector-strong  -O2 -D_FORTIFY_SOURCE=2 -Wformat -Wformat-security -Werror=format-security  --sysroot=/workdir/recipe-sysroot  -O2 -pipe -g -feliminate-unused-debug-types -fmacro-prefix-map=/workdir/git=/usr/src/debug/opencv/4.6.0-r0  -fdebug-prefix-map=/workdir/git=/usr/src/debug/opencv/4.6.0-r0  -fmacro-prefix-map=/workdir/build=/usr/src/debug/opencv/4.6.0-r0  -fdebug-prefix-map=/workdir/build=/usr/src/debug/opencv/4.6.0-r0  -fdebug-prefix-map=/workdir/recipe-sysroot=  -fmacro-prefix-map=/workdir/recipe-sysroot=  -fdebug-prefix-map=/workdir/recipe-sysroot-native=    -fsigned-char -W -Wall -Wreturn-type -Waddress -Wsequence-point -Wformat -Wformat-security -Wmissing-declarations -Wmissing-prototypes -Wstrict-prototypes -Wundef -Winit-self -Wpointer-arith -Wshadow -Wuninitialized -Wno-comment -Wimplicit-fallthrough=3 -Wno-strict-overflow -fdiagnostics-show-option -pthread -fomit-frame-pointer -ffunction-sections -fdata-sections  -DNDEBUG  -DNDEBUG\n"
"    C flags (Debug):             -mthumb -mfpu=neon -mfloat-abi=hard -mcpu=cortex-a9 -I/workdir/git/include  -fstack-protector-strong  -O2 -D_FORTIFY_SOURCE=2 -Wformat -Wformat-security -Werror=format-security  --sysroot=/workdir/recipe-sysroot  -O2 -pipe -g -feliminate-unused-debug-types -fmacro-prefix-map=/workdir/git=/usr/src/debug/opencv/4.6.0-r0  -fdebug-prefix-map=/workdir/git=/usr/src/debug/opencv/4.6.0-r0  -fmacro-prefix-map=/workdir/build=/usr/src/debug/opencv/4.6.0-r0  -fdebug-prefix-map=/workdir/build=/usr/src/debug/opencv/4.6.0-r0  -fdebug-prefix-map=/workdir/recipe-sysroot=  -fmacro-prefix-map=/workdir/recipe-sysroot=  -fdebug-prefix-map=/workdir/recipe-sysroot-native=    -fsigned-char -W -Wall -Wreturn-type -Waddress -Wsequence-point -Wformat -Wformat-security -Wmissing-declarations -Wmissing-prototypes -Wstrict-prototypes -Wundef -Winit-self -Wpointer-arith -Wshadow -Wuninitialized -Wno-comment -Wimplicit-fallthrough=3 -Wno-strict-overflow -fdiagnostics-show-option -pthread -fomit-frame-pointer -ffunction-sections -fdata-sections  -g  -DDEBUG -D_DEBUG\n"
"    Linker flags (Release):      -Wl,-O1 -Wl,--hash-style=gnu -Wl,--as-needed -fmacro-prefix-map=/workdir/git=/usr/src/debug/opencv/4.6.0-r0  -fdebug-prefix-map=/workdir/git=/usr/src/debug/opencv/4.6.0-r0  -fmacro-prefix-map=/workdir/build=/usr/src/debug/opencv/4.6.0-r0  -fdebug-prefix-map=/workdir/build=/usr/src/debug/opencv/4.6.0-r0  -fdebug-prefix-map=/workdir/recipe-sysroot=  -fmacro-prefix-map=/workdir/recipe-sysroot=  -fdebug-prefix-map=/workdir/recipe-sysroot-native=  -Wl,-z,relro,-z,now  -Wl,--gc-sections -Wl,--as-needed -Wl,--no-undefined  \n"
"    Linker flags (Debug):        -Wl,-O1 -Wl,--hash-style=gnu -Wl,--as-needed -fmacro-prefix-map=/workdir/git=/usr/src/debug/opencv/4.6.0-r0  -fdebug-prefix-map=/workdir/git=/usr/src/debug/opencv/4.6.0-r0  -fmacro-prefix-map=/workdir/build=/usr/src/debug/opencv/4.6.0-r0  -fdebug-prefix-map=/workdir/build=/usr/src/debug/opencv/4.6.0-r0  -fdebug-prefix-map=/workdir/recipe-sysroot=  -fmacro-prefix-map=/workdir/recipe-sysroot=  -fdebug-prefix-map=/workdir/recipe-sysroot-native=  -Wl,-z,relro,-z,now  -Wl,--gc-sections -Wl,--as-needed -Wl,--no-undefined  \n"
"    ccache:                      NO\n"
"    Precompiled headers:         NO\n"
"    Extra dependencies:          dl m pthread rt\n"
"    3rdparty dependencies:\n"
"\n"
"  OpenCV modules:\n"
"    To be built:                 alphamat aruco barcode bgsegm bioinspired calib3d ccalib core datasets dnn dnn_objdetect dnn_superres dpm face features2d flann fuzzy gapi hfs highgui img_hash imgcodecs imgproc intensity_transform line_descriptor mcc ml objdetect optflow phase_unwrapping photo plot python3 quality rapid reg rgbd saliency sfm shape stereo stitching structured_light superres surface_matching tracking ts video videoio videostab wechat_qrcode xfeatures2d ximgproc xobjdetect xphoto\n"
"    Disabled:                    freetype text world\n"
"    Disabled by dependency:      -\n"
"    Unavailable:                 cudaarithm cudabgsegm cudacodec cudafeatures2d cudafilters cudaimgproc cudalegacy cudaobjdetect cudaoptflow cudastereo cudawarping cudev cvv hdf java julia matlab ovis python2 viz\n"
"    Applications:                perf_tests examples apps\n"
"    Documentation:               NO\n"
"    Non-free algorithms:         NO\n"
"\n"
"  GUI:                           GTK3\n"
"    GTK+:                        YES (ver 3.24.34)\n"
"      GThread :                  YES (ver 2.72.3)\n"
"      GtkGlExt:                  NO\n"
"\n"
"  Media I/O: \n"
"    ZLib:                        /workdir/recipe-sysroot/usr/lib/libz.so (ver 1.2.13)\n"
"    JPEG:                        /workdir/recipe-sysroot/usr/lib/libjpeg.so (ver 62)\n"
"    WEBP:                        /workdir/recipe-sysroot/usr/lib/libwebp.so (ver encoder: 0x020f)\n"
"    PNG:                         /workdir/recipe-sysroot/usr/lib/libpng.so (ver 1.6.39)\n"
"    TIFF:                        /workdir/recipe-sysroot/usr/lib/libtiff.so (ver 42 / 4.4.0)\n"
"    JPEG 2000:                   build (ver 2.4.0)\n"
"    HDR:                         YES\n"
"    SUNRASTER:                   YES\n"
"    PXM:                         YES\n"
"    PFM:                         YES\n"
"\n"
"  Video I/O:\n"
"    GStreamer:                   YES (1.20.5)\n"
"    v4l/v4l2:                    YES (linux/videodev2.h)\n"
"    gPhoto2:                     YES\n"
"\n"
"  Parallel framework:            TBB (ver 2021.5 interface 12050)\n"
"\n"
"  Trace:                         YES (with Intel ITT)\n"
"\n"
"  Other third-party libraries:\n"
"    Lapack:                      NO\n"
"    Eigen:                       YES (ver 3.4.0)\n"
"    Custom HAL:                  YES (carotene (ver 0.0.1))\n"
"    Protobuf:                    /workdir/recipe-sysroot/usr/lib/libprotobuf.so (3.21.5)\n"
"\n"
"  Python 3:\n"
"    Interpreter:                 /workdir/recipe-sysroot-native/usr/bin/python3-native/python3 (ver 3.10.6)\n"
"    Libraries:                   /workdir/recipe-sysroot/usr/lib/libpython3.10.so (ver 3.10.6)\n"
"    numpy:                       /workdir/recipe-sysroot/usr/lib/python3.10/site-packages/numpy/core/include (ver undefined - cannot be probed because of the cross-compilation)\n"
"    install path:                lib/python3.10/site-packages/cv2/python-3.10\n"
"\n"
"  Python (for build):            /workdir/recipe-sysroot-native/usr/bin/python3-native/python3\n"
"\n"
"  Install to:                    /usr\n"
"-----------------------------------------------------------------\n"
"\n"
