/*M///////////////////////////////////////////////////////////////////////////////////////
//
//  IMPORTANT: READ BEFORE DOWNLOADING, COPYING, INSTALLING OR USING.
//
//  By downloading, copying, installing or using the software you agree to this license.
//  If you do not agree to this license, do not download, install,
//  copy or use the software.
//
//
//                           License Agreement
//                For Open Source Computer Vision Library
//
// Copyright (C) 2015, Itseez Inc, all rights reserved.
// Third party copyrights are property of their respective owners.
//
// Redistribution and use in source and binary forms, with or without modification,
// are permitted provided that the following conditions are met:
//
//   * Redistribution's of source code must retain the above copyright notice,
//     this list of conditions and the following disclaimer.
//
//   * Redistribution's in binary form must reproduce the above copyright notice,
//     this list of conditions and the following disclaimer in the documentation
//     and/or other materials provided with the distribution.
//
//   * The name of the copyright holders may not be used to endorse or promote products
//     derived from this software without specific prior written permission.
//
// This software is provided by the copyright holders and contributors "as is" and
// any express or implied warranties, including, but not limited to, the implied
// warranties of merchantability and fitness for a particular purpose are disclaimed.
// In no event shall the Itseez Inc or contributors be liable for any direct,
// indirect, incidental, special, exemplary, or consequential damages
// (including, but not limited to, procurement of substitute goods or services;
// loss of use, data, or profits; or business interruption) however caused
// and on any theory of liability, whether in contract, strict liability,
// or tort (including negligence or otherwise) arising in any way out of
// the use of this software, even if advised of the possibility of such damage.
//
//M*/

#include "dpm_convolution.hpp"

namespace cv
{
namespace dpm
{
double ConvolutionEngine::convolve(const Mat &feat, const Mat &filter,
        int dimHOG, int x, int y)
{
    double val = 0;
    for (int yp = 0; yp < filter.rows; yp++)
    {
        const double *pfeat = (double*)feat.ptr(y + yp) + x * dimHOG;
        const double *pfilter = (double*)filter.ptr(yp);

        for (int xp = 0; xp < filter.cols; xp++)
        {
            val += pfeat[xp] * pfilter[xp];
        }
    }

    return val;
}

void ConvolutionEngine::convolve(const Mat &feat, const Mat &filter,
        int dimHOG, Mat &result)
{
    for (int y = 0; y < result.rows; y++)
    {
        double *presult = (double*)result.ptr(y);
        for (int x = 0; x < result.cols; x++)
        {
            double val = 0;
            for (int yp = 0; yp < filter.rows; yp++)
            {
                const double *pfeat = (double*)feat.ptr(y + yp) + x * dimHOG;
                const double *pfilter = (double*)filter.ptr(yp);

                for (int xp = 0; xp < filter.cols; xp++)
                {
                    val += pfeat[xp] * pfilter[xp];
                }
            } // yp

            presult[x] = val;
        } // x
    } // y
}
} // namespace cv
} // namespace dpm
